<?php

/**
 * @file
 * Cron Support Functions for Eloqua
 *
 * @package Eloqua
 */

/**
 * Cron helper function.
 *
 * @param object $post
 *   Saved post.
 */
function _eloqua_cron($post = NULL) {
  // Ensure that cURL exists.  Otherwise this will fail hard.
  if (!function_exists('curl_init')) {
    $message = t('cURL does not appear to be installed.');
    $params = array();
    watchdog('eloqua', $message, $params, WATCHDOG_ALERT);
    return;
  }

  $posts = array();

  // If post passed into this function, only process this post.
  if ($post) {
    $posts = array($post);
  }
  // Post all the data to Eloqua and fetch unposted data.
  else {
    $result_set = db_select('eloqua_saved_posts')
      ->fields('eloqua_saved_posts')
      ->condition('status', ELOQUA_STATUS_NEW)
      ->range(0, (int) variable_get('batch_size', 50))
      ->execute()
      ->fetchAll();
    $posts = _eloqua_unserialize_data_column($result_set);
  }

  // If nothnig to do, or something funky happened, bail.
  if (empty($posts) || !is_array($posts)) {
    return;
  }

  foreach ($posts as $post) {
    if (!is_array($post->data)) {
      $post->data = array();
    }

    // Only pass along the accept-language header as a header. We pass the user
    // agent information in the cURL request below.
    $original_headers = $post->data['user_headers'];
    $headers = array(
      'Accept-Language' => array_key_exists('accept-language', $original_headers) ? $original_headers['accept-language'] : 'en',
    );

    // Fetch the post fields to send to Eloqua.
    $post_fields = _eloqua_cron_get_post_fields($post);

    // Create the cURL request.
    $options = array();
    $options['post_fields'] = $post_fields;
    $options['user_agent'] = array_key_exists('user-agent', $original_headers) ? $original_headers['user-agent'] : 'User Relay';
    $options['http_headers'] = $headers;
    $ch = _eloqua_cron_get_curl_resource($options);
    if (is_null($ch)) {
      $message = t('Something went wrong with cURL, unable to obtain handle.  Aborting status updates.');
      $params = array();
      watchdog('eloqua', $message, $params, WATCHDOG_ERROR);
      break;
    }
    try {
      $data = curl_exec($ch);
      $curl_status = curl_getinfo($ch);
      $status_class = (int) floor($curl_status['http_code'] / 100);
      $success = ($data) && ($status_class == 2 || $status_class == 3);
    }
    catch (Exception $e) {
      // PHP needs to implement finally.
      // No idea what went wrong, should just bail out of the process.
      // However, log.
      $message = t('Something went wrong with cURL. Uncaught exception: !message');
      $params = array(
        '!message' => $e->getMessage(),
      );
      watchdog('eloqua', $message, $params, WATCHDOG_ERROR);

      // Throwing an error here will result in an uncaught exception. This, when
      // it happens in cron, will result in the "Cron exceeding time limit"
      // error, because it didn't end properly. The above will log the message
      // to watchdog, and that's the extent of the notice that will be given. To
      // mimic current behavior, break will be used instead of continue.
      break;
    }
    curl_close($ch);

    // cURL.
    $post->data['server_post'][] = array(
      'timestamp' => time(),
      'response' => $data,
      'http_status' => $curl_status['http_code'],
    );

    // Update post data.
    if ($success) {
      $post->{'status'} = ELOQUA_STATUS_UPLOADED;
    }
    else {
      $post->{'status'} = ELOQUA_STATUS_FAILED;
    }
    eloqua_post_update($post);
  }
}

/**
 * Returns a configured cURL resource for use.
 *
 * @return resource
 *   The cURL resource.
 */
function _eloqua_cron_get_curl_resource($options = array()) {
  $ch = curl_init();
  if (empty($ch)) {
    $message = t('Unable to open cURL. Skipping message.');
    $params = array();
    watchdog('eloqua', $message, $params, WATCHDOG_ERROR);
    $ch = NULL;
  }

  if ($ch == NULL) {
    return NULL;
  }

  // Setting options.
  // TODO: Timeout admin settings.
  $curl_opts = array(
    CURLOPT_HEADER => 1,
    CURLOPT_POST => 1,
    CURLOPT_RETURNTRANSFER => 1,
    CURLOPT_CONNECTTIMEOUT => 5,
    CURLOPT_TIMEOUT => 5,
    CURLOPT_URL => 'http://now.eloqua.com/e/f2.aspx',
  );

  $option_map = array(
    'post_fields' => CURLOPT_POSTFIELDS,
    'user_agent' => CURLOPT_USERAGENT,
    'http_headers' => CURLOPT_HTTPHEADER,
  );
  foreach ($option_map as $key => $curl_key) {
    if (array_key_exists($key, $options)) {
      $curl_opts[$curl_key] = $options[$key];
    }
  }

  $status = _eloqua_set_curl_opts($ch, $curl_opts);
  $error_msg = _eloqua_set_curl_get_error($status);
  if (!empty($error_msg)) {
    // Failed setting an option.
    $message = t('Unable to set cURL options. Skipping message. !options.');
    $params = array(
      '!options' => '<pre>' . $error_msg . '</pre>',
    );
    watchdog('eloqua', $message, $params, WATCHDOG_ERROR);
    curl_close($ch);
    $ch = NULL;
  }
  return $ch;
}

/**
 * Handle translating post values into what Eloqua wants in terms of structure.
 *
 * Dates are by default handled in YYYY-mm-dd
 * Times are by default handled in HH:mm
 *
 * @param array $tree
 *   The post tree name => value pairs.
 * @param array $posted_values
 *   The post tree, could be name => value pairs or index => value pairs.
 * @param array $result
 *   The re-structured tree that Eloqua will leverage.
 */
function _eloqua_get_submission_data($tree, $posted_values, &$result) {
  foreach ($tree as $name => $value) {
    // Expand fieldsets.
    if (is_array($value) && !in_array($value, $posted_values)) {
      _eloqua_get_submission_data($value, $posted_values, $result);
    }
    // Convert multi-value fields into strings.
    elseif (is_array($value)) {
      $result[$name] = implode(',', $value);
    }
    // Everything else is just passed along.
    else {
      $result[$name] = $value;
    }
  }
}

/**
 * Remaps the post fields to be used within Eloqua.
 *
 * @param array $post_fields
 *   Data to post.
 * @param array $post
 *   Post data.
 *
 * @return array
 *   $post_fields modified to have remapped keys where defined.
 */
function _eloqua_cron_remap_post_fields($post_fields, $post) {
  $result = array();
  $nid = $post->data['form_post']['details']['nid'];
  if (empty($nid) || !is_numeric($nid)) {
    // Something is wrong with this.
    return $post_fields;
  }
  $node = node_load($nid);
  $components = $node->webform['components'];
  $map = array();
  if (is_array($components)) {
    foreach ($components as $component) {
      if (!empty($component['extra']['eloqua']['key'])
        && $component['extra']['eloqua']['key'] !== $component['form_key']) {
        $map[$component['form_key']] = $component['extra']['eloqua']['key'];
      }
    }
  }

  foreach ($post_fields as $key => $value) {
    if (isset($map[$key])) {
      $result[$map[$key]] = $value;
    }
    else {
      $result[$key] = $value;
    }
  }

  return $result;
}

/**
 * Returns the cURL error for a status returned from eloqua_curl_set_opts().
 *
 * @TODO: Theme
 *
 * @param array $status
 *   Status received from eloqua_curl_set_opts().
 *
 * @return string
 *   Html formatted message.
 *
 * @see eloqua_curl_set_opts()
 */
function _eloqua_set_curl_get_error($status) {
  if (!is_array($status)) {
    return FALSE;
  }
  $messages = array();
  $message = '';
  foreach ($status as $key => $data) {
    if ($data['status']) {
      continue;
    }
    $value = var_export($data['value'], TRUE);
    $messages[] = "<li>$key - Error: {$data['message']}<br /> Value: $value</li>";
  }
  if (!empty($messages)) {
    $message = '<ul>' . implode("\n", $messages) . '</ul>';
  }
  return $message;
}

/**
 * Sets cURL options en mass, checking for errors.
 *
 * @param resource $ch
 *   cURL resource.
 * @param array $options
 *   cURL settings to apply to the resource.
 *
 * @return array
 *   Status of all the set_opt() calls.
 */
function _eloqua_set_curl_opts(&$ch, $options = array()) {
  if (!is_array($options)) {
    return FALSE;
  }
  $result = array();
  foreach ($options as $key => $value) {
    $status = curl_setopt($ch, $key, $value);
    $status_msg = '';
    if (!$status) {
      $status_msg = curl_error($ch);
    }
    $result[$key] = array(
      'status' => $status,
      'value' => $value,
      'message' => $status_msg,
    );
  }
  return $result;
}

/**
 * Returns the post fields for the request, modified for Eloqua.
 *
 * @param object $post
 *   The post data.
 *
 * @return array
 *   The updated post data for Eloqua.
 */
function _eloqua_cron_get_post_fields($post) {
  // TODO: Review newest webform for exact behavior on whether submitted_tree
  // exists.
  $form_data = !empty($post->data['form_post']['submitted_tree']) ? $post->data['form_post']['submitted_tree'] : $post->data['form_post']['submitted'];

  $post_fields = array();
  // Merge any Eloqua values into the post to be submitted.
  foreach ($post->data['form_post'] as $name => $value) {
    if (substr($name, 0, 3) === 'elq') {
      $post_fields[$name] = $value;
    }
  }

  // Translates the data into a Eloqua-worthy format.
  _eloqua_get_submission_data($form_data, $post->data['form_post']['submitted'], $post_fields);

  // Remap any fields that can't be handled by webform.
  $post_fields = _eloqua_cron_remap_post_fields($post_fields, $post);
  return $post_fields;
}
